"""
Library to control the picomotor driver.
"""

import serial, time


class MSerial:
    channel_names = dict(ch1=0, ch2=1, ch3=2)
    driver_names = dict(driver1="a1", driver2="a2")
    type_names = dict(standard=0, tiny=1)
    
    def __init__(self, port, echo=True, max_retry=2, wait=0.1, sendwidget=None, recvwidget=None, **serial_kws):
        kws = dict(baudrate=19200, bytesize=serial.EIGHTBITS, 
                   parity=serial.PARITY_NONE, stopbits=serial.STOPBITS_ONE, 
                   timeout=0, xonxoff=True, rtscts=False, dsrdtr=False)
        kws.update(serial_kws)
        self.serial = serial.Serial(port, **kws)
        self.echo = echo
        self.wait = wait
        self.sendwidget = sendwidget
        self.recvwidget = recvwidget
        self.joystick_enable(enable=False)

    def send(self, cmd):
        """Send a command to the picomotor driver."""
        line = cmd + '\r\n'
        retval = self.serial.write(bytes(line, encoding='ascii'))
        self.serial.flush()
        if self.echo:
            self.log(cmd, widget=self.sendwidget)
        return retval
    
    def readlines(self):
        """Read response from picomotor driver."""
        return ''.join([l.decode('ASCII') for l in self.serial.readlines()])
    
    def log(self, msg, widget=None):
        if widget is None:
            print(msg, flush=True)
        else:
            widget.value = msg
        
    def sendrecv(self, cmd):
        """Send a command and (optionally) printing the picomotor driver's response."""
        res =self.send(cmd) 
        if self.echo:
            time.sleep(self.wait)
            ret_str = self.readlines()
            self.log(ret_str, widget=self.recvwidget)
        return res
    
    def set_channel(self, channel, driver, vel=None, acc=None):
        """Set current channel ('ch1', 'ch2' or 'ch3') and (optionally) its velocity."""
        assert channel in self.channel_names
        assert driver in self.driver_names
        fmt = dict(driver=self.driver_names[driver], channel=self.channel_names[channel])
        basecmd = '{cmd} {driver} {channel}={value}'
        if acc is not None:
            assert 0 < acc <= 32000, 'Acceleration out of range (1..32000).'
            cmd = basecmd.format(cmd='ACC', value=acc, **fmt)
            self.sendrecv(cmd)
        if vel is not None:
            assert 0 < vel <= 2000, 'Velocity out of range (1..2000).'
            cmd = basecmd.format(cmd='VEL', value=vel, **fmt)
            self.sendrecv(cmd)
        cmd = 'chl {driver}={channel}'.format(**fmt)
        return self.sendrecv(cmd)

    def set_type(self, channel, driver, typ):
        """Set picomotor type ('Standard' or 'Tiny') """
        assert channel in self.channel_names, 'Invalid value'
        assert driver in self.driver_names, 'Invalid value'
        assert typ in self.type_names, 'Invalid value'
        fmt = dict(driver=self.driver_names[driver], channel=self.channel_names[channel],
                   typ=self.type_names[typ])
        cmd = 'typ {driver} {channel}={typ}'.format(**fmt)
        return self.sendrecv(cmd)

    def move(self, steps, channel, driver, vel=None, acc=None, go=True):
        """Send command to move `channel` of the given `steps`."""
        self.set_channel(channel, driver, vel=vel, acc=acc)
        driver=self.driver_names[driver]
        cmd = 'rel {driver}={steps}'.format(driver=driver, steps=steps)
        if go:
            cmd = cmd + ' g'
        return self.sendrecv(cmd)

    def go(self):
        """Send 'go' command to execute all previously sent move commands."""
        return self.sendrecv('go')
    
    def halt(self):
        """Send 'HAL' command to stop motion with deceleration."""
        return self.sendrecv('hal')
    
    def joystick_enable(self, enable=True):
        """Enable or disable the joystick."""
        cmd = 'JON' if enable else 'JOF'
        return self.sendrecv(cmd)

    def status_msg(self):
        """Return the driver status byte as an integer (see manual pag. 185)."""
        self.send('STA')
        time.sleep(self.wait)
        ret_str = self.readlines()
        if self.echo:
            self.log(repr(ret_str), widget=self.recvwidget)
        return ret_str
    
    def status(self):
        ret_str = self.status_msg()
        i = ret_str.find('A1=')
        if i >= 0:
            status = int(ret_str[i+5:i+7], 16)
        else:
            raise IOError("Received: '%s'" % ret_str)
        return status
    
    def is_moving(self):
        """Return True if motor is moving, else False."""
        status = self.status()
        return status & 0x01